#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>

#define DEBUG

#undef ulong
#define ulong	unsigned long

#define NOACCESS	(1<<0)
#define NOCOOKIES	(1<<1)
#define NOAGENT		(1<<2)

typedef struct HostConf
{
	ulong	ip;
	ulong	ip_mask;
	int	deny;
	int	allow;

} HostConf;

HostConf hc[] =
{
{	0,		0,		NOCOOKIES|NOAGENT,	0		},	/* everything		*/
//{	0xD8200000,	0xFFFF0000,	NOACCESS,		0		},	/* yahoo		*/
{	0xD1CFE0DA,	0xFFFFFFFF,	NOACCESS,		0		},	/* ads.freshmeat.net	*/
{	0xD1CFE000,	0xFFFFFF00,	0,			NOCOOKIES	},	/* slashdot.org		*/
{	0,		0,		-1,			-1		},
};

static void sigchild(int sig)
{
	int	status;
	pid_t	pid;

	for(;;)
	{
		pid = waitpid(-1,&status,WNOHANG);
		if (pid <= 0)
			break;
	}
	signal(SIGCHLD,sigchild);
}

#define SL(x)	x,strlen(x)

struct	sockaddr_in sai;
struct	hostent *he;
struct	in_addr ia;
char	hln[8192];
char	hdr[8192];
char	tmp[8192];
char	*method,*url,*uri,*proto;
char	*p,*d;
char	*host;
ulong	ip,xip;
int	port,httpd;
int	n,lf,skip;
int	blocking;

static void gethostflags(void)
{
	blocking = 0;
	ip = ntohl(ia.s_addr);
	for(n=0;(hc[n].deny >= 0);n++)
	{
		xip = (ip & hc[n].ip_mask);
		if (xip == hc[n].ip)
		{
			blocking |= hc[n].deny;
			blocking &= ~hc[n].allow;
		}
	}
}

static void headerparse(void)
{
	int	hashost = 0;

	/* header blocking */
	p = hdr;
	d = tmp;
	lf = 1;
	skip = 0;
	while(*p)
	{
		if (*p == '\n')
		{
			lf = 2;
			if (skip)
			{
				skip = 0;
				p++;
				continue;
			}
		}
		else
		if (lf)
			lf--;
		if (lf == 1)
		{
			if (!strncasecmp(p,SL("Host:")))
				hashost = 1;

			if ((blocking & NOAGENT) && (!strncasecmp(p,SL("User-Agent:"))))
				skip = 1;
			else
			if ((blocking & NOCOOKIES) && (!strncasecmp(p,SL("Cookie:"))))
				skip = 1;
		}
		if (skip)
			p++;
		else
			*(d++) = *(p++);
	}
	*d = 0;
	d = hdr;
	if (!hashost)
	{
		sprintf(d,"Host: %s\n",url);
		d = strchr(d,0);
	}
	strcpy(d,tmp);
}

#define output(s,x)	write(s,x,strlen(x))

static void proxy(int s)
{
	p = hln;
	lf = 0;
	for(;;)
	{
		if (read(s,p,1) <= 0)
			break;
		if (*p == '\r')
			continue;
		if (*p == '\n')
		{
			*(++p) = 0;
			if (lf == 0)
			{
				lf = 2;
				p = hdr;
				continue;
			}
			if (lf == 2)
			{
				break;
			}
			lf = 2;
			continue;
		}
		else
		{
			lf = (lf) ? 1 : 0;
		}
		p++;
	}
#ifdef DEBUG
	fprintf(stderr,"[%.4X] hln = %s\n",getpid(),hln);
//	fprintf(stderr,"[%.4X] hdr = '%s'\n",getpid(),hdr);
	fflush(stderr);
#endif
	method = hln;
	url = strchr(hln,' ');
	*(url++) = 0;
	proto = strchr(url,' ');
	*(proto++) = 0;
	if ((p = strchr(proto,'\r')))
		*p = 0;
	if ((p = strchr(proto,'\n')))
		*p = 0;
	if (strncasecmp(url,"http://",7))
	{
#ifdef DEBUG
		fprintf(stderr,"proto mismatch\n");
#endif
		close(s);
		exit(0);
	}
	url += 7;
	uri = strchr(url,'/');
	if (uri)
	{
		*(uri++) = 0;
	}
	else
	{
		uri = "";
	}
	port = 80;
	host = strchr(url,':');
	if (host)
	{
		*(host++) = 0;
		port = 0;
		while(*host)
		{
			port = (port * 10) + (*host - '0');
			host++;
		}
#ifdef DEBUG
		fprintf(stderr,"port = %i\n",port);
#endif
		if ((port < 1) || (port > 65535))
		{
#ifdef DEBUG
			fprintf(stderr,"erroneous port number\n");
#endif
			exit(1);
		}
	}
#ifdef DEBUG
	fprintf(stderr,"[%.4X] %s %s /%s %s\n",getpid(),method,url,uri,proto);
#endif
	host = url;
	he = gethostbyname(host);
	if (!he)
	{
		p =	"HTTP/1.0 500 Error\r\n"
			"Content-Type: text/html\r\n"
			"Connection: Close\r\n"
			"\r\n"
			"<html><head><title>Unable to resolve hostname</title></head>\r\n"
			"<body><h2>ERROR</h2>\r\n"
			"Unable to resolve hostname "
			;
		write(s,p,strlen(p));
#ifdef DEBUG
		fprintf(stderr,"%s",p);
#endif
		write(s,host,strlen(host));
#ifdef DEBUG
		fprintf(stderr,"%s",host);
#endif
		p =	"<br>\r\n"
			"</body></html>\r\n"
			;
		write(s,p,strlen(p));
#ifdef DEBUG
		fprintf(stderr,"%s",p);
#endif
		close(s);
		exit(0);
	}
	memcpy(&ia,he->h_addr_list[0],sizeof(ia));
#ifdef DEBUG
	fprintf(stderr,"[%.4X] %s --> %s\n",getpid(),host,inet_ntoa(ia));
#endif

	gethostflags();
	headerparse();

	if (blocking & NOACCESS)
	{
#ifdef DEBUG
		fprintf(stderr,"access denied\n");
#endif
		output(s,"HTTP/1.0 403 Forbidden\r\n");
		output(s,"Content-Type: text/plain\r\n");
		output(s,"Connection: Close\r\n");
		output(s,"\r\n");
		output(s,"Error: Access to this object is denied\r\n");
		close(s);
		exit(0);
	}

#ifdef DEBUG
	fprintf(stderr,"new hdr = \n'\n%s'\n",hdr);
#endif

	memset(&sai,0,sizeof(sai));
	httpd = socket(AF_INET,SOCK_STREAM,0);
	sai.sin_family = AF_INET;
	sai.sin_port = htons(port);
	sai.sin_addr.s_addr = ia.s_addr;
	connect(httpd,(struct sockaddr*)&sai,sizeof(sai));
	sprintf(tmp,"%s /%s %s\r\n",method,uri,proto);
	write(httpd,tmp,strlen(tmp));
#ifdef DEBUG
	fprintf(stderr,"[%.4X] --> %s",getpid(),tmp);
#endif
	fcntl(s,F_SETFL,O_RDWR|O_NONBLOCK);
	fcntl(httpd,F_SETFL,O_RDWR|O_NONBLOCK);
	write(httpd,hdr,strlen(hdr));
	for(;;)
	{
		if (s >= 0)
		{
			n = read(s,tmp,sizeof(tmp));
			if ((n == 0) || ((n < 0) && (errno != EAGAIN)))
			{
#ifdef DEBUG
				fprintf(stderr,"[%.4X] client closed connection\n",getpid());
#endif
				close(s);
				s = -1;
			}
			else
			if (n > 0)
			{
				write(httpd,tmp,n);
#ifdef DEBUG
				fprintf(stderr,"[%.4X] --> ... (%i)\n",getpid(),n);
#endif
			}
		}
		if (httpd >= 0)
		{
			n = read(httpd,tmp,sizeof(tmp));
			if ((n == 0) || ((n < 0) && (errno != EAGAIN)))
			{
#ifdef DEBUG
				fprintf(stderr,"[%.4X] server closed connection\n",getpid());
#endif
				close(httpd);
				httpd = -1;
				close(s);
				s = -1;
			}
			else
			if (n > 0)
			{
				write(s,tmp,n);
#ifdef DEBUG
				fprintf(stderr,"[%.4X] <-- ... (%i)\n",getpid(),n);
#endif
			}
		}
		if ((s < 0) && (httpd < 0))
		{
#ifdef DEBUG
			fprintf(stderr,"[%.4X] proxy done\n",getpid());
#endif
			exit(0);
		}
	}
}

int main(int argc, char **argv)
{
	struct	sockaddr_in sai;
	int	s,client,sz;

	signal(SIGCHLD,sigchild);
	switch(fork())
	{
	case 0:
		break;
	default:
		exit(0);
	}
	if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
		exit(1);
	memset(&sai,0,sizeof(sai));
	sai.sin_family = AF_INET;
	sai.sin_port = htons(8080);
	sai.sin_addr.s_addr = INADDR_ANY;
	if (bind(s,(struct sockaddr*)&sai,sizeof(sai)) < 0)
		exit(1);
	if (listen(s,4) < 0)
		exit(1);
	for(;;)
	{
		sz = sizeof(sai);
		client = accept(s,(struct sockaddr*)&sai,&sz);
		switch(fork())
		{
		case -1:
			exit(1);
		case 0:
			close(s);
			proxy(client);
			exit(0);
		default:
			close(client);
		}
	}
}
